/* ****************************************************** */
/*    Wilson fermion solver                               */
/*                                                        */
/*    This program is written in C-like manner, but       */
/*   using C++ functions.  Please compile by g++, or      */
/*   other C++ compiler.  [18 Apr 2009 HM]                */
/*                                                        */
/*                     Copyright(c) Hideo Matsufuru 2009  */
/* ****************************************************** */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "lattsize.h"

#include "cg_wilson.h"


/* ****************************************************** */
void solve_CG(double enorm, int &nconv, double &diff,
                       double *xq, double *u, double *b){

  static double x[Nvst], s[Nvst], r[Nvst], p[Nvst];
  // double x[Nvst], s[Nvst], r[Nvst], p[Nvst];

  int niter = 1000;

  equate(Nvst, s, b);
  double snorm, sr;
  normv(Nvst, sr, s);
  snorm = 1.0/sr;

  double rr, rrp;
  nconv = -1;

  solve_CG_init(rrp, rr, u, x, r, s, p, s);
    printf("  init: %22.15e\n", rr*snorm);

  for(int iter=0; iter < niter; iter++){

    solve_CG_step(rrp, rr, u, x, r, s, p, s);

       printf("%6d  %22.15e\n", iter, rr*snorm);

    if(rr*snorm < enorm){
      nconv = iter;
      break;
    }

  }

  if(nconv == -1){
    printf(" not converged\n");
  }

  //  printf("converged:\n");
  //  printf("  nconv = %d\n", nconv);

  equate(Nvst,xq,x);

  opr_DdagD(r,u,x);
  selfadd(Nvst,r,b,-1.0);
  normv(Nvst,diff,r);
  
}
/* ****************************************************** */
void solve_CG_init(double &rrp, double &rr, double *u,
     double *x, double *r, double *s, double *p, double *v){

  equate(Nvst,r,s);
  equate(Nvst,x,s);

  opr_DdagD(s,u,x);

  selfadd(Nvst,r,s,-1.0);
  equate(Nvst,p,r);

  double rrt;
  normv(Nvst,rrt,r);
  rr = rrt;
  rrp = rr;

}
/* ****************************************************** */
void solve_CG_step(double &rrp, double &rr, double *u,
     double *x, double *r, double *s, double *p, double *v){

  double cr, bk, pap;

  opr_DdagD(v,u,p);

  vecprd(Nvst,pap,v,p);
  cr = rrp/pap;

  selfadd(Nvst,x,p,cr);

  selfadd(Nvst,r,v,-cr);
  double rrt;
  normv(Nvst,rrt,r);
  rr=rrt;
  bk = rr/rrp;

  selfsprd(Nvst,p,bk);
  selfadd(Nvst,p,r,1.0);

  rrp = rr;

}
/* ****************************************************** */
void opr_DdagD(double *v, double *u, double *w){

  static double v2[Nvst];

  opr_H(v2,u,w );
  opr_H(v ,u,v2);

}
/* ****************************************************** */
void opr_D(double *v, double *u, double *w){

  setconst(Nvst, v, 0.0);

  mult_xp(v,u,w);
  mult_xm(v,u,w);
  mult_yp(v,u,w);
  mult_ym(v,u,w);
  mult_zp(v,u,w);
  mult_zm(v,u,w);
  mult_tp(v,u,w);
  mult_tm(v,u,w);

  spx(Nvst,v,w,-CKs);

}
/* ****************************************************** */
void opr_H(double *v, double *u, double *w){

  setconst(Nvst, v, 0.0);

  mult_xp(v,u,w);
  mult_xm(v,u,w);
  mult_yp(v,u,w);
  mult_ym(v,u,w);
  mult_zp(v,u,w);
  mult_zm(v,u,w);
  mult_tp(v,u,w);
  mult_tm(v,u,w);
  int id1 = 0;
  int id2 = Nvc;
  int id3 = Nvc*2;
  int id4 = Nvc*3;

  for(int ist=0; ist < Nst; ist++){
    int iv = ist*Nvc*ND;
   for(int ivc=0; ivc < Nvc; ivc++){
     double v3 = v[ivc+id3+iv];
     double v4 = v[ivc+id4+iv];
     v[ivc+id3+iv] = w[ivc+id1+iv] - CKs * v[ivc+id1+iv];
     v[ivc+id4+iv] = w[ivc+id2+iv] - CKs * v[ivc+id2+iv];
     v[ivc+id1+iv] = w[ivc+id3+iv] - CKs * v3;
     v[ivc+id2+iv] = w[ivc+id4+iv] - CKs * v4;
   }
  }

}
/* ****************************************************** */
void mult_xp(double *v2, double *u, double *v1){

  int idir = 0;

  int id1 = 0;
  int id2 = Nvc;
  int id3 = Nvc*2;
  int id4 = Nvc*3;

  int ix, nn;

  static double vt1[Nvc], vt2[Nvc];
  double wt1r, wt1i, wt2r, wt2i;

  // boundary part 
    ix = Nx-1;
    nn = 0;
    for(int iyzt=0; iyzt < (Ny*Nz*Nt); iyzt++){

    int iv = Nvc*ND*(ix + iyzt*Nx);
    int in = Nvc*ND*(nn + iyzt*Nx);
    int ig = Ndf*(ix + iyzt*Nx + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = v1[2*ic   +id1+in] - v1[2*ic+1 +id4+in];
      vt1[2*ic+1] = v1[2*ic+1 +id1+in] + v1[2*ic   +id4+in];
      vt2[2*ic  ] = v1[2*ic   +id2+in] - v1[2*ic+1 +id3+in];
      vt2[2*ic+1] = v1[2*ic+1 +id2+in] + v1[2*ic   +id3+in];
    }

    for(int ic=0; ic < Ncol; ic++){
      int ic2 = ic*Nvc;

      wt1r = u[0+ic2+ig]*vt1[0] - u[1+ic2+ig]*vt1[1]
           + u[2+ic2+ig]*vt1[2] - u[3+ic2+ig]*vt1[3]
           + u[4+ic2+ig]*vt1[4] - u[5+ic2+ig]*vt1[5];
      wt1i = u[0+ic2+ig]*vt1[1] + u[1+ic2+ig]*vt1[0]
           + u[2+ic2+ig]*vt1[3] + u[3+ic2+ig]*vt1[2]
           + u[4+ic2+ig]*vt1[5] + u[5+ic2+ig]*vt1[4];

      wt2r = u[0+ic2+ig]*vt2[0] - u[1+ic2+ig]*vt2[1]
           + u[2+ic2+ig]*vt2[2] - u[3+ic2+ig]*vt2[3]
           + u[4+ic2+ig]*vt2[4] - u[5+ic2+ig]*vt2[5];
      wt2i = u[0+ic2+ig]*vt2[1] + u[1+ic2+ig]*vt2[0]
           + u[2+ic2+ig]*vt2[3] + u[3+ic2+ig]*vt2[2]
           + u[4+ic2+ig]*vt2[5] + u[5+ic2+ig]*vt2[4];

      v2[2*ic   +id1+iv] =  wt1r;
      v2[2*ic+1 +id1+iv] =  wt1i;
      v2[2*ic   +id2+iv] =  wt2r;
      v2[2*ic+1 +id2+iv] =  wt2i;
      v2[2*ic   +id3+iv] =  wt2i;
      v2[2*ic+1 +id3+iv] = -wt2r;
      v2[2*ic   +id4+iv] =  wt1i;
      v2[2*ic+1 +id4+iv] = -wt1r;
    }
  }

  // bulk part 
  for(int iyzt=0; iyzt < (Ny*Nz*Nt); iyzt++){
   for(int ix=0; ix < (Nx-1); ix++){
      nn = ix+1;

    int iv = Nvc*ND*(ix + iyzt*Nx);
    int in = Nvc*ND*(nn + iyzt*Nx);
    int ig = Ndf*(ix + iyzt*Nx + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = v1[2*ic   +id1+in] - v1[2*ic+1 +id4+in];
      vt1[2*ic+1] = v1[2*ic+1 +id1+in] + v1[2*ic   +id4+in];
      vt2[2*ic  ] = v1[2*ic   +id2+in] - v1[2*ic+1 +id3+in];
      vt2[2*ic+1] = v1[2*ic+1 +id2+in] + v1[2*ic   +id3+in];
    }

    for(int ic=0; ic < Ncol; ic++){
      int ic2 = ic*Nvc;

      wt1r = u[0+ic2+ig]*vt1[0] - u[1+ic2+ig]*vt1[1]
           + u[2+ic2+ig]*vt1[2] - u[3+ic2+ig]*vt1[3]
           + u[4+ic2+ig]*vt1[4] - u[5+ic2+ig]*vt1[5];
      wt1i = u[0+ic2+ig]*vt1[1] + u[1+ic2+ig]*vt1[0]
           + u[2+ic2+ig]*vt1[3] + u[3+ic2+ig]*vt1[2]
           + u[4+ic2+ig]*vt1[5] + u[5+ic2+ig]*vt1[4];

      wt2r = u[0+ic2+ig]*vt2[0] - u[1+ic2+ig]*vt2[1]
           + u[2+ic2+ig]*vt2[2] - u[3+ic2+ig]*vt2[3]
           + u[4+ic2+ig]*vt2[4] - u[5+ic2+ig]*vt2[5];
      wt2i = u[0+ic2+ig]*vt2[1] + u[1+ic2+ig]*vt2[0]
           + u[2+ic2+ig]*vt2[3] + u[3+ic2+ig]*vt2[2]
           + u[4+ic2+ig]*vt2[5] + u[5+ic2+ig]*vt2[4];

      v2[2*ic   +id1+iv] =  wt1r;
      v2[2*ic+1 +id1+iv] =  wt1i;
      v2[2*ic   +id2+iv] =  wt2r;
      v2[2*ic+1 +id2+iv] =  wt2i;
      v2[2*ic   +id3+iv] =  wt2i;
      v2[2*ic+1 +id3+iv] = -wt2r;
      v2[2*ic   +id4+iv] =  wt1i;
      v2[2*ic+1 +id4+iv] = -wt1r;
    }
   }
  }

}

/* ****************************************************** */
void mult_xm(double *v2, double *u, double *v1){

  int idir = 0;

  int id1 = 0;
  int id2 = Nvc;
  int id3 = Nvc*2;
  int id4 = Nvc*3;

  int ix, nn;

  double vt1[Nvc], vt2[Nvc];
  double wt1r, wt1i, wt2r, wt2i;


  // boundary part 
    ix = 0;
    nn = Nx-1;
    for(int iyzt=0; iyzt<(Ny*Nz*Nt); iyzt++){

    int iv = Nvc*ND*(ix + iyzt*Nx);
    int in = Nvc*ND*(nn + iyzt*Nx);
    int ig = Ndf*(nn + iyzt*Nx + idir*Nst);

    for(int ic=0; ic<Ncol; ic++){
      vt1[2*ic  ] = v1[2*ic   +id1+in] + v1[2*ic+1 +id4+in];
      vt1[2*ic+1] = v1[2*ic+1 +id1+in] - v1[2*ic   +id4+in];

      vt2[2*ic  ] = v1[2*ic   +id2+in] + v1[2*ic+1 +id3+in];
      vt2[2*ic+1] = v1[2*ic+1 +id2+in] - v1[2*ic   +id3+in];
    }

    for(int ic=0; ic < Ncol; ic++){
      int icr = 2*ic;
      int ici = 2*ic+1;
      int ic1 = 0;
      int ic2 = Nvc;
      int ic3 = 2*Nvc;

      wt1r = u[icr+ic1+ig]*vt1[0] + u[ici+ic1+ig]*vt1[1]
           + u[icr+ic2+ig]*vt1[2] + u[ici+ic2+ig]*vt1[3]
           + u[icr+ic3+ig]*vt1[4] + u[ici+ic3+ig]*vt1[5];
      wt1i = u[icr+ic1+ig]*vt1[1] - u[ici+ic1+ig]*vt1[0]
           + u[icr+ic2+ig]*vt1[3] - u[ici+ic2+ig]*vt1[2]
           + u[icr+ic3+ig]*vt1[5] - u[ici+ic3+ig]*vt1[4];

      wt2r = u[icr+ic1+ig]*vt2[0] + u[ici+ic1+ig]*vt2[1]
           + u[icr+ic2+ig]*vt2[2] + u[ici+ic2+ig]*vt2[3]
           + u[icr+ic3+ig]*vt2[4] + u[ici+ic3+ig]*vt2[5];
      wt2i = u[icr+ic1+ig]*vt2[1] - u[ici+ic1+ig]*vt2[0]
           + u[icr+ic2+ig]*vt2[3] - u[ici+ic2+ig]*vt2[2]
           + u[icr+ic3+ig]*vt2[5] - u[ici+ic3+ig]*vt2[4];

      v2[icr +id1+iv] +=  wt1r;
      v2[ici +id1+iv] +=  wt1i;
      v2[icr +id2+iv] +=  wt2r;
      v2[ici +id2+iv] +=  wt2i;
      v2[icr +id3+iv] += -wt2i;
      v2[ici +id3+iv] += +wt2r;
      v2[icr +id4+iv] += -wt1i;
      v2[ici +id4+iv] += +wt1r;
    }
  }

  // bulk part 
  for(int iyzt=0; iyzt<Ny*Nz*Nt; iyzt++){
   for(int ix=1; ix<Nx; ix++){
      nn = ix-1;

    int iv = Nvc*ND*(ix + (iyzt)*Nx);
    int in = Nvc*ND*(nn + (iyzt)*Nx);
    int ig = Ndf*(nn + iyzt*Nx + idir*Nst);

    for(int ic=0; ic<Ncol; ic++){
      vt1[2*ic  ] = v1[2*ic   +id1+in] + v1[2*ic+1 +id4+in];
      vt1[2*ic+1] = v1[2*ic+1 +id1+in] - v1[2*ic   +id4+in];

      vt2[2*ic  ] = v1[2*ic   +id2+in] + v1[2*ic+1 +id3+in];
      vt2[2*ic+1] = v1[2*ic+1 +id2+in] - v1[2*ic   +id3+in];
    }

    for(int ic=0; ic<Ncol; ic++){
      int icr = 2*ic;
      int ici = 2*ic+1;
      int ic1 = 0;
      int ic2 = Nvc;
      int ic3 = 2*Nvc;

      wt1r = u[icr+ic1+ig]*vt1[0] + u[ici+ic1+ig]*vt1[1]
           + u[icr+ic2+ig]*vt1[2] + u[ici+ic2+ig]*vt1[3]
           + u[icr+ic3+ig]*vt1[4] + u[ici+ic3+ig]*vt1[5];
      wt1i = u[icr+ic1+ig]*vt1[1] - u[ici+ic1+ig]*vt1[0]
           + u[icr+ic2+ig]*vt1[3] - u[ici+ic2+ig]*vt1[2]
           + u[icr+ic3+ig]*vt1[5] - u[ici+ic3+ig]*vt1[4];

      wt2r = u[icr+ic1+ig]*vt2[0] + u[ici+ic1+ig]*vt2[1]
           + u[icr+ic2+ig]*vt2[2] + u[ici+ic2+ig]*vt2[3]
           + u[icr+ic3+ig]*vt2[4] + u[ici+ic3+ig]*vt2[5];
      wt2i = u[icr+ic1+ig]*vt2[1] - u[ici+ic1+ig]*vt2[0]
           + u[icr+ic2+ig]*vt2[3] - u[ici+ic2+ig]*vt2[2]
           + u[icr+ic3+ig]*vt2[5] - u[ici+ic3+ig]*vt2[4];

      v2[icr +id1+iv] +=  wt1r;
      v2[ici +id1+iv] +=  wt1i;
      v2[icr +id2+iv] +=  wt2r;
      v2[ici +id2+iv] +=  wt2i;
      v2[icr +id3+iv] += -wt2i;
      v2[ici +id3+iv] += +wt2r;
      v2[icr +id4+iv] += -wt1i;
      v2[ici +id4+iv] += +wt1r;
    }
   }
  }


}
/* ****************************************************** */
void mult_yp(double *v2, double *u, double *v1){

  int idir = 1;

  int id1 = 0;
  int id2 = Nvc;
  int id3 = Nvc*2;
  int id4 = Nvc*3;

  static double vt1[Nvc], vt2[Nvc];
  double wt1r, wt1i, wt2r, wt2i;

  // boundary part 
    int iy = Ny-1;
    int nn = 0;
   for(int izt=0; izt < (Nz*Nt); izt++){
    for(int ix=0; ix < Nx; ix++){

    int iv = Nvc*ND*(ix + iy*Nx + izt*Nx*Ny);
    int in = Nvc*ND*(ix + nn*Nx + izt*Nx*Ny);
    int ig = Ndf*(ix + iy*Nx + izt*Nx*Ny + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = v1[2*ic   +id1+in] + v1[2*ic   +id4+in];
      vt1[2*ic+1] = v1[2*ic+1 +id1+in] + v1[2*ic+1 +id4+in];
      vt2[2*ic  ] = v1[2*ic   +id2+in] - v1[2*ic   +id3+in];
      vt2[2*ic+1] = v1[2*ic+1 +id2+in] - v1[2*ic+1 +id3+in];
    }

    for(int ic=0; ic < Ncol; ic++){
      int ic2 = ic*Nvc;

      wt1r = u[0+ic2+ig]*vt1[0] - u[1+ic2+ig]*vt1[1]
           + u[2+ic2+ig]*vt1[2] - u[3+ic2+ig]*vt1[3]
           + u[4+ic2+ig]*vt1[4] - u[5+ic2+ig]*vt1[5];
      wt1i = u[0+ic2+ig]*vt1[1] + u[1+ic2+ig]*vt1[0]
           + u[2+ic2+ig]*vt1[3] + u[3+ic2+ig]*vt1[2]
           + u[4+ic2+ig]*vt1[5] + u[5+ic2+ig]*vt1[4];

      wt2r = u[0+ic2+ig]*vt2[0] - u[1+ic2+ig]*vt2[1]
           + u[2+ic2+ig]*vt2[2] - u[3+ic2+ig]*vt2[3]
           + u[4+ic2+ig]*vt2[4] - u[5+ic2+ig]*vt2[5];
      wt2i = u[0+ic2+ig]*vt2[1] + u[1+ic2+ig]*vt2[0]
           + u[2+ic2+ig]*vt2[3] + u[3+ic2+ig]*vt2[2]
           + u[4+ic2+ig]*vt2[5] + u[5+ic2+ig]*vt2[4];

      v2[2*ic   +id1+iv] +=  wt1r;
      v2[2*ic+1 +id1+iv] +=  wt1i;
      v2[2*ic   +id2+iv] +=  wt2r;
      v2[2*ic+1 +id2+iv] +=  wt2i;
      v2[2*ic   +id3+iv] += -wt2r;
      v2[2*ic+1 +id3+iv] += -wt2i;
      v2[2*ic   +id4+iv] +=  wt1r;
      v2[2*ic+1 +id4+iv] +=  wt1i;
    }
   }
  }

  // bulk part 
  for(int izt=0; izt < (Nz*Nt); izt++){
   for(int iy=0; iy < (Ny-1); iy++){
     int nn = iy + 1;
    for(int ix=0; ix < Nx; ix++){

    int iv = Nvc*ND*(ix + iy*Nx + izt*Nx*Ny);
    int in = Nvc*ND*(ix + nn*Nx + izt*Nx*Ny);
    int ig = Ndf*(ix + iy*Nx + izt*Nx*Ny + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = v1[2*ic   +id1+in] + v1[2*ic   +id4+in];
      vt1[2*ic+1] = v1[2*ic+1 +id1+in] + v1[2*ic+1 +id4+in];
      vt2[2*ic  ] = v1[2*ic   +id2+in] - v1[2*ic   +id3+in];
      vt2[2*ic+1] = v1[2*ic+1 +id2+in] - v1[2*ic+1 +id3+in];
    }

    for(int ic=0; ic < Ncol; ic++){
      int ic2 = ic*Nvc;

      wt1r = u[0+ic2+ig]*vt1[0] - u[1+ic2+ig]*vt1[1]
           + u[2+ic2+ig]*vt1[2] - u[3+ic2+ig]*vt1[3]
           + u[4+ic2+ig]*vt1[4] - u[5+ic2+ig]*vt1[5];
      wt1i = u[0+ic2+ig]*vt1[1] + u[1+ic2+ig]*vt1[0]
           + u[2+ic2+ig]*vt1[3] + u[3+ic2+ig]*vt1[2]
           + u[4+ic2+ig]*vt1[5] + u[5+ic2+ig]*vt1[4];

      wt2r = u[0+ic2+ig]*vt2[0] - u[1+ic2+ig]*vt2[1]
           + u[2+ic2+ig]*vt2[2] - u[3+ic2+ig]*vt2[3]
           + u[4+ic2+ig]*vt2[4] - u[5+ic2+ig]*vt2[5];
      wt2i = u[0+ic2+ig]*vt2[1] + u[1+ic2+ig]*vt2[0]
           + u[2+ic2+ig]*vt2[3] + u[3+ic2+ig]*vt2[2]
           + u[4+ic2+ig]*vt2[5] + u[5+ic2+ig]*vt2[4];

      v2[2*ic   +id1+iv] +=  wt1r;
      v2[2*ic+1 +id1+iv] +=  wt1i;
      v2[2*ic   +id2+iv] +=  wt2r;
      v2[2*ic+1 +id2+iv] +=  wt2i;
      v2[2*ic   +id3+iv] += -wt2r;
      v2[2*ic+1 +id3+iv] += -wt2i;
      v2[2*ic   +id4+iv] +=  wt1r;
      v2[2*ic+1 +id4+iv] +=  wt1i;
     }
    }
   }
  }

}

/* ****************************************************** */
void mult_ym(double *v2, double *u, double *v1){

  int idir = 1;

  int id1 = 0;
  int id2 = Nvc;
  int id3 = Nvc*2;
  int id4 = Nvc*3;

  double vt1[Nvc], vt2[Nvc];
  double wt1r, wt1i, wt2r, wt2i;


  // boundary part 
    int iy = 0;
    int nn = Ny-1;
   for(int izt=0; izt < (Nz*Nt); izt++){
    for(int ix=0; ix < Nx; ix++){

    int iv = Nvc*ND*(ix + iy*Nx + izt*Nx*Ny);
    int in = Nvc*ND*(ix + nn*Nx + izt*Nx*Ny);
    int ig = Ndf*(ix + nn*Nx + izt*Nx*Ny + idir*Nst);

    for(int ic=0; ic<Ncol; ic++){
      vt1[2*ic  ] = v1[2*ic   +id1+in] - v1[2*ic   +id4+in];
      vt1[2*ic+1] = v1[2*ic+1 +id1+in] - v1[2*ic+1 +id4+in];

      vt2[2*ic  ] = v1[2*ic   +id2+in] + v1[2*ic   +id3+in];
      vt2[2*ic+1] = v1[2*ic+1 +id2+in] + v1[2*ic+1 +id3+in];
    }

    for(int ic=0; ic < Ncol; ic++){
      int icr = 2*ic;
      int ici = 2*ic+1;
      int ic1 = 0;
      int ic2 = Nvc;
      int ic3 = 2*Nvc;

      wt1r = u[icr+ic1+ig]*vt1[0] + u[ici+ic1+ig]*vt1[1]
           + u[icr+ic2+ig]*vt1[2] + u[ici+ic2+ig]*vt1[3]
           + u[icr+ic3+ig]*vt1[4] + u[ici+ic3+ig]*vt1[5];
      wt1i = u[icr+ic1+ig]*vt1[1] - u[ici+ic1+ig]*vt1[0]
           + u[icr+ic2+ig]*vt1[3] - u[ici+ic2+ig]*vt1[2]
           + u[icr+ic3+ig]*vt1[5] - u[ici+ic3+ig]*vt1[4];

      wt2r = u[icr+ic1+ig]*vt2[0] + u[ici+ic1+ig]*vt2[1]
           + u[icr+ic2+ig]*vt2[2] + u[ici+ic2+ig]*vt2[3]
           + u[icr+ic3+ig]*vt2[4] + u[ici+ic3+ig]*vt2[5];
      wt2i = u[icr+ic1+ig]*vt2[1] - u[ici+ic1+ig]*vt2[0]
           + u[icr+ic2+ig]*vt2[3] - u[ici+ic2+ig]*vt2[2]
           + u[icr+ic3+ig]*vt2[5] - u[ici+ic3+ig]*vt2[4];

      v2[icr +id1+iv] +=  wt1r;
      v2[ici +id1+iv] +=  wt1i;
      v2[icr +id2+iv] +=  wt2r;
      v2[ici +id2+iv] +=  wt2i;
      v2[icr +id3+iv] +=  wt2r;
      v2[ici +id3+iv] +=  wt2i;
      v2[icr +id4+iv] += -wt1r;
      v2[ici +id4+iv] += -wt1i;
    }
   }
  }

  // bulk part 
  for(int izt=0; izt < (Nz*Nt); izt++){
   for(int iy=1; iy < Ny; iy++){
     int nn = iy - 1;
    for(int ix=0; ix < Nx; ix++){

    int iv = Nvc*ND*(ix + iy*Nx + izt*Nx*Ny);
    int in = Nvc*ND*(ix + nn*Nx + izt*Nx*Ny);
    int ig = Ndf*(ix + nn*Nx + izt*Nx*Ny + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = v1[2*ic   +id1+in] - v1[2*ic   +id4+in];
      vt1[2*ic+1] = v1[2*ic+1 +id1+in] - v1[2*ic+1 +id4+in];

      vt2[2*ic  ] = v1[2*ic   +id2+in] + v1[2*ic   +id3+in];
      vt2[2*ic+1] = v1[2*ic+1 +id2+in] + v1[2*ic+1 +id3+in];
    }

    for(int ic=0; ic<Ncol; ic++){
      int icr = 2*ic;
      int ici = 2*ic+1;
      int ic1 = 0;
      int ic2 = Nvc;
      int ic3 = 2*Nvc;

      wt1r = u[icr+ic1+ig]*vt1[0] + u[ici+ic1+ig]*vt1[1]
           + u[icr+ic2+ig]*vt1[2] + u[ici+ic2+ig]*vt1[3]
           + u[icr+ic3+ig]*vt1[4] + u[ici+ic3+ig]*vt1[5];
      wt1i = u[icr+ic1+ig]*vt1[1] - u[ici+ic1+ig]*vt1[0]
           + u[icr+ic2+ig]*vt1[3] - u[ici+ic2+ig]*vt1[2]
           + u[icr+ic3+ig]*vt1[5] - u[ici+ic3+ig]*vt1[4];

      wt2r = u[icr+ic1+ig]*vt2[0] + u[ici+ic1+ig]*vt2[1]
           + u[icr+ic2+ig]*vt2[2] + u[ici+ic2+ig]*vt2[3]
           + u[icr+ic3+ig]*vt2[4] + u[ici+ic3+ig]*vt2[5];
      wt2i = u[icr+ic1+ig]*vt2[1] - u[ici+ic1+ig]*vt2[0]
           + u[icr+ic2+ig]*vt2[3] - u[ici+ic2+ig]*vt2[2]
           + u[icr+ic3+ig]*vt2[5] - u[ici+ic3+ig]*vt2[4];

      v2[icr +id1+iv] +=  wt1r;
      v2[ici +id1+iv] +=  wt1i;
      v2[icr +id2+iv] +=  wt2r;
      v2[ici +id2+iv] +=  wt2i;
      v2[icr +id3+iv] +=  wt2r;
      v2[ici +id3+iv] +=  wt2i;
      v2[icr +id4+iv] += -wt1r;
      v2[ici +id4+iv] += -wt1i;
     }
    }
   }
  }


}
/* ****************************************************** */
void mult_zp(double *v2, double *u, double *v1){

  int idir = 2;

  int id1 = 0;
  int id2 = Nvc;
  int id3 = Nvc*2;
  int id4 = Nvc*3;

  static double vt1[Nvc], vt2[Nvc];
  double wt1r, wt1i, wt2r, wt2i;

  // boundary part 
    int iz = Nz-1;
    int nn = 0;
   for(int it=0; it < Nt; it++){
     for(int ixy=0; ixy < (Nx*Ny); ixy++){

    int iv = Nvc*ND*(ixy + iz*Nx*Ny + it*Nx*Ny*Nz);
    int in = Nvc*ND*(ixy + nn*Nx*Ny + it*Nx*Ny*Nz);
    int ig = Ndf*(ixy + iz*Nx*Ny + it*Nx*Ny*Nz + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = v1[2*ic   +id1+in] - v1[2*ic+1 +id3+in];
      vt1[2*ic+1] = v1[2*ic+1 +id1+in] + v1[2*ic   +id3+in];
      vt2[2*ic  ] = v1[2*ic   +id2+in] + v1[2*ic+1 +id4+in];
      vt2[2*ic+1] = v1[2*ic+1 +id2+in] - v1[2*ic   +id4+in];
    }

    for(int ic=0; ic < Ncol; ic++){
      int ic2 = ic*Nvc;

      wt1r = u[0+ic2+ig]*vt1[0] - u[1+ic2+ig]*vt1[1]
           + u[2+ic2+ig]*vt1[2] - u[3+ic2+ig]*vt1[3]
           + u[4+ic2+ig]*vt1[4] - u[5+ic2+ig]*vt1[5];
      wt1i = u[0+ic2+ig]*vt1[1] + u[1+ic2+ig]*vt1[0]
           + u[2+ic2+ig]*vt1[3] + u[3+ic2+ig]*vt1[2]
           + u[4+ic2+ig]*vt1[5] + u[5+ic2+ig]*vt1[4];

      wt2r = u[0+ic2+ig]*vt2[0] - u[1+ic2+ig]*vt2[1]
           + u[2+ic2+ig]*vt2[2] - u[3+ic2+ig]*vt2[3]
           + u[4+ic2+ig]*vt2[4] - u[5+ic2+ig]*vt2[5];
      wt2i = u[0+ic2+ig]*vt2[1] + u[1+ic2+ig]*vt2[0]
           + u[2+ic2+ig]*vt2[3] + u[3+ic2+ig]*vt2[2]
           + u[4+ic2+ig]*vt2[5] + u[5+ic2+ig]*vt2[4];

      v2[2*ic   +id1+iv] +=  wt1r;
      v2[2*ic+1 +id1+iv] +=  wt1i;
      v2[2*ic   +id2+iv] +=  wt2r;
      v2[2*ic+1 +id2+iv] +=  wt2i;
      v2[2*ic   +id3+iv] +=  wt1i;
      v2[2*ic+1 +id3+iv] += -wt1r;
      v2[2*ic   +id4+iv] += -wt2i;
      v2[2*ic+1 +id4+iv] +=  wt2r;
    }
   }
  }

  // bulk part 
  for(int it=0; it < Nt; it++){
   for(int iz=0; iz < (Nz-1); iz++){
      int nn = iz + 1;
    for(int ixy=0; ixy < (Nx*Ny); ixy++){

    int iv = Nvc*ND*(ixy + iz*Nx*Ny + it*Nx*Ny*Nz);
    int in = Nvc*ND*(ixy + nn*Nx*Ny + it*Nx*Ny*Nz);
    int ig = Ndf*(ixy + iz*Nx*Ny + it*Nx*Ny*Nz + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = v1[2*ic   +id1+in] - v1[2*ic+1 +id3+in];
      vt1[2*ic+1] = v1[2*ic+1 +id1+in] + v1[2*ic   +id3+in];
      vt2[2*ic  ] = v1[2*ic   +id2+in] + v1[2*ic+1 +id4+in];
      vt2[2*ic+1] = v1[2*ic+1 +id2+in] - v1[2*ic   +id4+in];
    }

    for(int ic=0; ic < Ncol; ic++){
      int ic2 = ic*Nvc;

      wt1r = u[0+ic2+ig]*vt1[0] - u[1+ic2+ig]*vt1[1]
           + u[2+ic2+ig]*vt1[2] - u[3+ic2+ig]*vt1[3]
           + u[4+ic2+ig]*vt1[4] - u[5+ic2+ig]*vt1[5];
      wt1i = u[0+ic2+ig]*vt1[1] + u[1+ic2+ig]*vt1[0]
           + u[2+ic2+ig]*vt1[3] + u[3+ic2+ig]*vt1[2]
           + u[4+ic2+ig]*vt1[5] + u[5+ic2+ig]*vt1[4];

      wt2r = u[0+ic2+ig]*vt2[0] - u[1+ic2+ig]*vt2[1]
           + u[2+ic2+ig]*vt2[2] - u[3+ic2+ig]*vt2[3]
           + u[4+ic2+ig]*vt2[4] - u[5+ic2+ig]*vt2[5];
      wt2i = u[0+ic2+ig]*vt2[1] + u[1+ic2+ig]*vt2[0]
           + u[2+ic2+ig]*vt2[3] + u[3+ic2+ig]*vt2[2]
           + u[4+ic2+ig]*vt2[5] + u[5+ic2+ig]*vt2[4];

      v2[2*ic   +id1+iv] +=  wt1r;
      v2[2*ic+1 +id1+iv] +=  wt1i;
      v2[2*ic   +id2+iv] +=  wt2r;
      v2[2*ic+1 +id2+iv] +=  wt2i;
      v2[2*ic   +id3+iv] +=  wt1i;
      v2[2*ic+1 +id3+iv] += -wt1r;
      v2[2*ic   +id4+iv] += -wt2i;
      v2[2*ic+1 +id4+iv] +=  wt2r;
     }
    }
   }
  }

}
/* ****************************************************** */
void mult_zm(double *v2, double *u, double *v1){

  int idir = 2;

  int id1 = 0;
  int id2 = Nvc;
  int id3 = Nvc*2;
  int id4 = Nvc*3;

  double vt1[Nvc], vt2[Nvc];
  double wt1r, wt1i, wt2r, wt2i;


  // boundary part 
    int iz = 0;
    int nn = Nz-1;
   for(int it=0; it < Nt; it++){
    for(int ixy=0; ixy < (Nx*Ny); ixy++){

    int iv = Nvc*ND*(ixy + iz*Nx*Ny + it*Nx*Ny*Nz);
    int in = Nvc*ND*(ixy + nn*Nx*Ny + it*Nx*Ny*Nz);
    int ig = Ndf*(ixy + nn*Nx*Ny + it*Nx*Ny*Nz + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = v1[2*ic   +id1+in] + v1[2*ic+1 +id3+in];
      vt1[2*ic+1] = v1[2*ic+1 +id1+in] - v1[2*ic   +id3+in];
      vt2[2*ic  ] = v1[2*ic   +id2+in] - v1[2*ic+1 +id4+in];
      vt2[2*ic+1] = v1[2*ic+1 +id2+in] + v1[2*ic   +id4+in];
    }

    for(int ic=0; ic < Ncol; ic++){
      int icr = 2*ic;
      int ici = 2*ic+1;
      int ic1 = 0;
      int ic2 = Nvc;
      int ic3 = 2*Nvc;

      wt1r = u[icr+ic1+ig]*vt1[0] + u[ici+ic1+ig]*vt1[1]
           + u[icr+ic2+ig]*vt1[2] + u[ici+ic2+ig]*vt1[3]
           + u[icr+ic3+ig]*vt1[4] + u[ici+ic3+ig]*vt1[5];
      wt1i = u[icr+ic1+ig]*vt1[1] - u[ici+ic1+ig]*vt1[0]
           + u[icr+ic2+ig]*vt1[3] - u[ici+ic2+ig]*vt1[2]
           + u[icr+ic3+ig]*vt1[5] - u[ici+ic3+ig]*vt1[4];

      wt2r = u[icr+ic1+ig]*vt2[0] + u[ici+ic1+ig]*vt2[1]
           + u[icr+ic2+ig]*vt2[2] + u[ici+ic2+ig]*vt2[3]
           + u[icr+ic3+ig]*vt2[4] + u[ici+ic3+ig]*vt2[5];
      wt2i = u[icr+ic1+ig]*vt2[1] - u[ici+ic1+ig]*vt2[0]
           + u[icr+ic2+ig]*vt2[3] - u[ici+ic2+ig]*vt2[2]
           + u[icr+ic3+ig]*vt2[5] - u[ici+ic3+ig]*vt2[4];

      v2[icr +id1+iv] +=  wt1r;
      v2[ici +id1+iv] +=  wt1i;
      v2[icr +id2+iv] +=  wt2r;
      v2[ici +id2+iv] +=  wt2i;
      v2[icr +id3+iv] += -wt1i;
      v2[ici +id3+iv] +=  wt1r;
      v2[icr +id4+iv] +=  wt2i;
      v2[ici +id4+iv] += -wt2r;
    }
   }
  }

  // bulk part 
  for(int it=0; it < Nt; it++){
   for(int iz=1; iz < Nz; iz++){
      int nn = iz - 1;
    for(int ixy=0; ixy < (Nx*Ny); ixy++){

    int iv = Nvc*ND*(ixy + iz*Nx*Ny + it*Nx*Ny*Nz);
    int in = Nvc*ND*(ixy + nn*Nx*Ny + it*Nx*Ny*Nz);
    int ig = Ndf*(ixy + nn*Nx*Ny + it*Nx*Ny*Nz + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = v1[2*ic   +id1+in] + v1[2*ic+1 +id3+in];
      vt1[2*ic+1] = v1[2*ic+1 +id1+in] - v1[2*ic   +id3+in];
      vt2[2*ic  ] = v1[2*ic   +id2+in] - v1[2*ic+1 +id4+in];
      vt2[2*ic+1] = v1[2*ic+1 +id2+in] + v1[2*ic   +id4+in];
    }

    for(int ic=0; ic<Ncol; ic++){
      int icr = 2*ic;
      int ici = 2*ic+1;
      int ic1 = 0;
      int ic2 = Nvc;
      int ic3 = 2*Nvc;

      wt1r = u[icr+ic1+ig]*vt1[0] + u[ici+ic1+ig]*vt1[1]
           + u[icr+ic2+ig]*vt1[2] + u[ici+ic2+ig]*vt1[3]
           + u[icr+ic3+ig]*vt1[4] + u[ici+ic3+ig]*vt1[5];
      wt1i = u[icr+ic1+ig]*vt1[1] - u[ici+ic1+ig]*vt1[0]
           + u[icr+ic2+ig]*vt1[3] - u[ici+ic2+ig]*vt1[2]
           + u[icr+ic3+ig]*vt1[5] - u[ici+ic3+ig]*vt1[4];

      wt2r = u[icr+ic1+ig]*vt2[0] + u[ici+ic1+ig]*vt2[1]
           + u[icr+ic2+ig]*vt2[2] + u[ici+ic2+ig]*vt2[3]
           + u[icr+ic3+ig]*vt2[4] + u[ici+ic3+ig]*vt2[5];
      wt2i = u[icr+ic1+ig]*vt2[1] - u[ici+ic1+ig]*vt2[0]
           + u[icr+ic2+ig]*vt2[3] - u[ici+ic2+ig]*vt2[2]
           + u[icr+ic3+ig]*vt2[5] - u[ici+ic3+ig]*vt2[4];

      v2[icr +id1+iv] +=  wt1r;
      v2[ici +id1+iv] +=  wt1i;
      v2[icr +id2+iv] +=  wt2r;
      v2[ici +id2+iv] +=  wt2i;
      v2[icr +id3+iv] += -wt1i;
      v2[ici +id3+iv] +=  wt1r;
      v2[icr +id4+iv] +=  wt2i;
      v2[ici +id4+iv] += -wt2r;
     }
    }
   }
  }

}
/* ****************************************************** */
void mult_tp(double *v2, double *u, double *v1){

  int idir = 3;

  //  int id1 = 0;
  //  int id2 = Nvc;
  int id3 = Nvc*2;
  int id4 = Nvc*3;

  static double vt1[Nvc], vt2[Nvc];
  double wt1r, wt1i, wt2r, wt2i;

  // boundary part 
    int it = Nt-1;
    int nn = 0;
   for(int ixyz=0; ixyz < (Nx*Ny*Nz); ixyz++){

    int iv = Nvc*ND*(ixyz + it*Nx*Ny*Nz);
    int in = Nvc*ND*(ixyz + nn*Nx*Ny*Nz);
    int ig = Ndf*(ixyz + it*Nx*Ny*Nz + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = 2.0 * v1[2*ic   +id3+in];
      vt1[2*ic+1] = 2.0 * v1[2*ic+1 +id3+in];
      vt2[2*ic  ] = 2.0 * v1[2*ic   +id4+in];
      vt2[2*ic+1] = 2.0 * v1[2*ic+1 +id4+in];
    }

    for(int ic=0; ic < Ncol; ic++){
      int ic2 = ic*Nvc;

      wt1r = u[0+ic2+ig]*vt1[0] - u[1+ic2+ig]*vt1[1]
           + u[2+ic2+ig]*vt1[2] - u[3+ic2+ig]*vt1[3]
           + u[4+ic2+ig]*vt1[4] - u[5+ic2+ig]*vt1[5];
      wt1i = u[0+ic2+ig]*vt1[1] + u[1+ic2+ig]*vt1[0]
           + u[2+ic2+ig]*vt1[3] + u[3+ic2+ig]*vt1[2]
           + u[4+ic2+ig]*vt1[5] + u[5+ic2+ig]*vt1[4];

      wt2r = u[0+ic2+ig]*vt2[0] - u[1+ic2+ig]*vt2[1]
           + u[2+ic2+ig]*vt2[2] - u[3+ic2+ig]*vt2[3]
           + u[4+ic2+ig]*vt2[4] - u[5+ic2+ig]*vt2[5];
      wt2i = u[0+ic2+ig]*vt2[1] + u[1+ic2+ig]*vt2[0]
           + u[2+ic2+ig]*vt2[3] + u[3+ic2+ig]*vt2[2]
           + u[4+ic2+ig]*vt2[5] + u[5+ic2+ig]*vt2[4];

      v2[2*ic   +id3+iv] +=  wt1r;
      v2[2*ic+1 +id3+iv] +=  wt1i;
      v2[2*ic   +id4+iv] +=  wt2r;
      v2[2*ic+1 +id4+iv] +=  wt2i;
    }
  }

  // bulk part 
  for(int it=0; it < (Nt-1); it++){
      int nn = it + 1;
   for(int ixyz=0; ixyz < (Nx*Ny*Nz); ixyz++){

    int iv = Nvc*ND*(ixyz + it*Nx*Ny*Nz);
    int in = Nvc*ND*(ixyz + nn*Nx*Ny*Nz);
    int ig = Ndf*(ixyz + it*Nx*Ny*Nz + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = 2.0 * v1[2*ic   +id3+in];
      vt1[2*ic+1] = 2.0 * v1[2*ic+1 +id3+in];
      vt2[2*ic  ] = 2.0 * v1[2*ic   +id4+in];
      vt2[2*ic+1] = 2.0 * v1[2*ic+1 +id4+in];
    }

    for(int ic=0; ic < Ncol; ic++){
      int ic2 = ic*Nvc;

      wt1r = u[0+ic2+ig]*vt1[0] - u[1+ic2+ig]*vt1[1]
           + u[2+ic2+ig]*vt1[2] - u[3+ic2+ig]*vt1[3]
           + u[4+ic2+ig]*vt1[4] - u[5+ic2+ig]*vt1[5];
      wt1i = u[0+ic2+ig]*vt1[1] + u[1+ic2+ig]*vt1[0]
           + u[2+ic2+ig]*vt1[3] + u[3+ic2+ig]*vt1[2]
           + u[4+ic2+ig]*vt1[5] + u[5+ic2+ig]*vt1[4];

      wt2r = u[0+ic2+ig]*vt2[0] - u[1+ic2+ig]*vt2[1]
           + u[2+ic2+ig]*vt2[2] - u[3+ic2+ig]*vt2[3]
           + u[4+ic2+ig]*vt2[4] - u[5+ic2+ig]*vt2[5];
      wt2i = u[0+ic2+ig]*vt2[1] + u[1+ic2+ig]*vt2[0]
           + u[2+ic2+ig]*vt2[3] + u[3+ic2+ig]*vt2[2]
           + u[4+ic2+ig]*vt2[5] + u[5+ic2+ig]*vt2[4];

      v2[2*ic   +id3+iv] +=  wt1r;
      v2[2*ic+1 +id3+iv] +=  wt1i;
      v2[2*ic   +id4+iv] +=  wt2r;
      v2[2*ic+1 +id4+iv] +=  wt2i;
    }
   }
  }

}
/* ****************************************************** */
void mult_tm(double *v2, double *u, double *v1){

  int idir = 3;

  int id1 = 0;
  int id2 = Nvc;
  // int id3 = Nvc*2;
  // int id4 = Nvc*3;

  double vt1[Nvc], vt2[Nvc];
  double wt1r, wt1i, wt2r, wt2i;


  // boundary part 
    int it = 0;
    int nn = Nt-1;
   for(int ixyz=0; ixyz < (Nx*Ny*Nz); ixyz++){

    int iv = Nvc*ND*(ixyz + it*Nx*Ny*Nz);
    int in = Nvc*ND*(ixyz + nn*Nx*Ny*Nz);
    int ig = Ndf*(ixyz + nn*Nx*Ny*Nz + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = 2.0 * v1[2*ic   +id1+in];
      vt1[2*ic+1] = 2.0 * v1[2*ic+1 +id1+in];
      vt2[2*ic  ] = 2.0 * v1[2*ic   +id2+in];
      vt2[2*ic+1] = 2.0 * v1[2*ic+1 +id2+in];
    }

    for(int ic=0; ic < Ncol; ic++){
      int icr = 2*ic;
      int ici = 2*ic+1;
      int ic1 = 0;
      int ic2 = Nvc;
      int ic3 = 2*Nvc;

      wt1r = u[icr+ic1+ig]*vt1[0] + u[ici+ic1+ig]*vt1[1]
           + u[icr+ic2+ig]*vt1[2] + u[ici+ic2+ig]*vt1[3]
           + u[icr+ic3+ig]*vt1[4] + u[ici+ic3+ig]*vt1[5];
      wt1i = u[icr+ic1+ig]*vt1[1] - u[ici+ic1+ig]*vt1[0]
           + u[icr+ic2+ig]*vt1[3] - u[ici+ic2+ig]*vt1[2]
           + u[icr+ic3+ig]*vt1[5] - u[ici+ic3+ig]*vt1[4];

      wt2r = u[icr+ic1+ig]*vt2[0] + u[ici+ic1+ig]*vt2[1]
           + u[icr+ic2+ig]*vt2[2] + u[ici+ic2+ig]*vt2[3]
           + u[icr+ic3+ig]*vt2[4] + u[ici+ic3+ig]*vt2[5];
      wt2i = u[icr+ic1+ig]*vt2[1] - u[ici+ic1+ig]*vt2[0]
           + u[icr+ic2+ig]*vt2[3] - u[ici+ic2+ig]*vt2[2]
           + u[icr+ic3+ig]*vt2[5] - u[ici+ic3+ig]*vt2[4];

      v2[icr +id1+iv] +=  wt1r;
      v2[ici +id1+iv] +=  wt1i;
      v2[icr +id2+iv] +=  wt2r;
      v2[ici +id2+iv] +=  wt2i;
    }
  }

  // bulk part 
  for(int it=1; it < Nt; it++){
      int nn = it - 1;
   for(int ixyz=0; ixyz < (Nx*Ny*Nz); ixyz++){

    int iv = Nvc*ND*(ixyz + it*Nx*Ny*Nz);
    int in = Nvc*ND*(ixyz + nn*Nx*Ny*Nz);
    int ig = Ndf*(ixyz + nn*Nx*Ny*Nz + idir*Nst);

    for(int ic=0; ic < Ncol; ic++){
      vt1[2*ic  ] = 2.0 * v1[2*ic   +id1+in];
      vt1[2*ic+1] = 2.0 * v1[2*ic+1 +id1+in];
      vt2[2*ic  ] = 2.0 * v1[2*ic   +id2+in];
      vt2[2*ic+1] = 2.0 * v1[2*ic+1 +id2+in];
    }

    for(int ic=0; ic<Ncol; ic++){
      int icr = 2*ic;
      int ici = 2*ic+1;
      int ic1 = 0;
      int ic2 = Nvc;
      int ic3 = 2*Nvc;

      wt1r = u[icr+ic1+ig]*vt1[0] + u[ici+ic1+ig]*vt1[1]
           + u[icr+ic2+ig]*vt1[2] + u[ici+ic2+ig]*vt1[3]
           + u[icr+ic3+ig]*vt1[4] + u[ici+ic3+ig]*vt1[5];
      wt1i = u[icr+ic1+ig]*vt1[1] - u[ici+ic1+ig]*vt1[0]
           + u[icr+ic2+ig]*vt1[3] - u[ici+ic2+ig]*vt1[2]
           + u[icr+ic3+ig]*vt1[5] - u[ici+ic3+ig]*vt1[4];

      wt2r = u[icr+ic1+ig]*vt2[0] + u[ici+ic1+ig]*vt2[1]
           + u[icr+ic2+ig]*vt2[2] + u[ici+ic2+ig]*vt2[3]
           + u[icr+ic3+ig]*vt2[4] + u[ici+ic3+ig]*vt2[5];
      wt2i = u[icr+ic1+ig]*vt2[1] - u[ici+ic1+ig]*vt2[0]
           + u[icr+ic2+ig]*vt2[3] - u[ici+ic2+ig]*vt2[2]
           + u[icr+ic3+ig]*vt2[5] - u[ici+ic3+ig]*vt2[4];

      v2[icr +id1+iv] +=  wt1r;
      v2[ici +id1+iv] +=  wt1i;
      v2[icr +id2+iv] +=  wt2r;
      v2[ici +id2+iv] +=  wt2i;
     }
   }
  }

}
/* ****************************************************** */
void set_src(int ic, int id,
              int ix, int iy, int iz, int it, double *v){

  setconst(Nvst, v, 0.0);

  int i = 2*ic + id*Nvc + Nvc*ND*(ix + iy*Nszx
                        + iz*Nszx*Nszy + it*Nszx*Nszy*Nszz);

  v[i]=1.0;

}
/* ****************************************************** */
void setconst(int nv, double *v, double a){

  for(int i=0; i < nv; i++){
    v[i] = a;
  }

}
/* ****************************************************** */
void vecprd(int nv, double &a, double *v1, double *v2){

  a = 0.0;
  for(int i=0; i < nv; i++){
    a = a + v1[i]*v2[i];
  }

}
/* ****************************************************** */
void normv(int nv, double &a, double *v){

  a = 0.0;
  for(int i=0; i < nv; i++){
    a = a + v[i]*v[i];
  }

}
/* ****************************************************** */
void selfadd(int nv, double *v, double *w, double a){

  for(int i=0; i < nv; i++){
    v[i] = v[i] + a * w[i];
  }

}
/* ****************************************************** */
void selfsprd(int nv, double *v, double a){

  for(int i=0; i < nv; i++){
    v[i] = a * v[i];
  }

}
/* ****************************************************** */
void spx(int nv, double *v,  double *w, double a){

  for(int i=0; i < nv; i++){
    v[i] = a * v[i] + w[i];
  }

}
/* ****************************************************** */
void equate(int nv, double *v,  double *w){

  for(int i=0; i < nv; i++){
    v[i] = w[i];
  }

}
/* ****************************************************** */
